from collections import namedtuple

from ..detectors.haar_cascade_detector import HaarCascadeDetector
from ..detectors.threshold_detector import ThresholdDetector

GroupToDetector = namedtuple("GroupToDetector", "constructor arguments")
_GROUP_TO_DETECTOR = {"threshold": GroupToDetector(ThresholdDetector, "threshold amd/or morphological functions"),
                      "haar_cascade": GroupToDetector(HaarCascadeDetector, "haar_classifier")}


def _argument_group_to_detector_constructor(argument_group, *args, **kwargs):
    return _GROUP_TO_DETECTOR[argument_group].constructor(*args, **kwargs)


def arguments_to_detector(mutually_exclusive_arguments):
    existing_group = None
    existing_group_name = None
    for group_name, argument_group in mutually_exclusive_arguments:
        if any(argument_group):
            if existing_group_name:
                raise ValueError(
                    "When passing parameters that create a Detector only 1 group can be passed,"
                    " got both '{}' and '{}'".format(_GROUP_TO_DETECTOR[group_name].arguments,
                                                     _GROUP_TO_DETECTOR[argument_group].arguments))
            else:
                existing_group = argument_group
                existing_group_name = group_name

    if existing_group_name is None:
        detector = None
    elif existing_group_name != "detector":
        detector = _argument_group_to_detector_constructor(existing_group_name, *existing_group)
    else:
        detector = existing_group
    return detector



from .network_location import NetworkLocation


class Connection:
    def send(self, data, *args, **kwargs):
        """
        Used to send information to an external source,
        send represents outgoing information, usually over-network but not exclusively

        :param data: the data to be sent, a common use is sending the directions from direction objects after
        object detection
        :param args: any additional arguments for the specific connection object
        :param kwargs: any additional keyword arguments for the specific connection object
        :return:
        """
        pass

    def receive(self, *args, **kwargs):
        """
        Used to receive information from an external source,
        receive represents incoming information, usually over-network but not exclusively
        Receive can be used to send a request in order to receive
        for example HttpConnection send an http request and returns the response data

        :param args: any additional arguments for the specific connection object
        :param kwargs: any additional keyword arguments for the specific connection object
        :return:
        """
        pass

    def receive_from_location(self, network_location: NetworkLocation, *args, **kwargs):
        """
        Used to receive information from an external source,
        receive represents incoming information, usually over-network but not exclusively

        Receive can be used to send a request in order to receive
        for example HttpConnection send an http request and returns the response data

        Receive_from_location uses a NetworkLocation to specify

        :param network_location: used to specify parameters of a specific "location"
        or information about the receipt.
        :param args: any additional arguments for the specific connection object
        :param kwargs: any additional keyword arguments for the specific connection object
        :return:
        """
        return self.receive(*args, **network_location, **kwargs,)

    def send_to_location(self, data, network_location: NetworkLocation, *args, **kwargs):
        return self.receive(*args, **network_location, **kwargs, data=data)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        pass



from typing import *
import math

from ..direction_modifiers.direction_modifier import DirectionModifier
from ..camera_.camera_settings import *


class Director:
    """
    Directors are object that extract usable information from the objects that were detected image.
    This information be where they are located in the image relative to the center (directions for a robot),
    how many objects were detected, identifying an object or any other logical usage of the object detection
    rather than raw image processing or computer vision.
    This is the final step before sending/using the data.

    Director.direct works in the following way:
    - if the amount of contours is less than target amount then set directions to failed_detection
    - else it passes only target_amount of contours to the direction function
    - The final result (direction function or failed_detection) is passed to the direction monitors
    - each monitor is applied to the result and its result is passed to the next one
    - then the final result is returned
    """

    def __init__(self, directing_function: Callable, failed_detection: Any = None, target_amount: int = 1,
                 direction_modifiers: List[DirectionModifier] = None):
        """
        :param directing_function: the function that performs the initial direction calculation
        :param failed_detection: The value returned on a failed detection
        :param target_amount: the target amount to find, if the value is 0
                              the target amount (minimum amount) removed and is infinite
        :param direction_modifiers: list of DirectionMonitor objects that alter the directions to be sent
        """
        self.directions = directing_function
        if not isinstance(target_amount, int):
            raise ValueError("Target amount must be an integer, got '{}' of type '{}'.".format(target_amount,
                                                                                               type(target_amount)))
        self.target_amount = target_amount if target_amount != 0 else math.inf
        self.direction_monitors = direction_modifiers
        self.failed_detection = failed_detection

    def direct(self, contours: List[np.ndarray], image: np.ndarray, sorter=None) -> Any:
        """
        Returns the directions using the vision.director for the given contours using the image and camera settings (if
        given).

        Director.direct works in the following way:
            - if the amount of contours is less than target amount then set directions to failed_detection
            - else it passes only target_amount of contours to the direction function
            - The final result (direction function or failed_detection) is passed to the direction monitors
            - each monitor is applied to the result and its result is passed to the next one
            - then the final result is returned

        :param contours: the list of contours (numpy ndarrays) from which to extrapolate target direction
        :param image: the image from which the contours were taken
        :param sorter: A sorter function to be applied on the contours, or None to not apply one
        :return: Depends on the directing function and the direction monitors,
                 usually a number or a string
        """
        if not len(contours) >= self.target_amount:
            directions = self.failed_detection
        else:
            if sorter is not None:
                contours = sorter(contours)
            if math.isfinite(self.target_amount):
                contours = contours[:self.target_amount]
            directions = self.directions(contours, image)

        return self.apply_direction_monitors(directions, contours, image)

    def apply_direction_monitors(self, directions: Any, contours: List[np.ndarray],
                                 image: np.ndarray) -> Any:
        """
        Applies the list of direction monitors one after the other

        Direction monitors are functions that receive directions contours image and camera settings
        and change it some way, like applying a PID feedback loop, returning a "stop" value based on some condition
        (like getting close enough to the target)

        :param directions: the raw directions returned from the directions function
        :param contours: the list of contours that passed all filters
        :param image: the image from which the contours were found
        :return: the final direction result
        """
        if self.direction_monitors:
            for direction_monitor in self.direction_monitors:
                directions = direction_monitor.monitor(directions, contours, image)
                if direction_monitor.priority:
                    return directions
        return directions



import typing

from .connection import Connection


class SerialConnection(Connection):
    """
    Serial connection is a Connection used to send and receive on a Serial bus like various USB cables.

    SerialConnection is implemented under the hood using Pyserial and it is a package requirement for
    using SerialConnection, make sure it is installed by executing in console:

    pip install pyserial

    """
    def __init__(self, port: typing.Union[int, str], baudrate: int = 9600, *args, **kwargs):
        from serial import Serial

        self.socket = Serial(port, baudrate, *args, **kwargs)

    def close(self):
        self.socket.close()

    def send(self, data, *args, **kwargs):
        self.socket.write(data)

    def receive(self, *args, **kwargs):
        return self.socket.read(*args, **kwargs)



import copy
import math

from .display_image import *
from ..image_utilities.open_image import open_image


def display_contours(image, contours, amount=0, display_loop=False, window_name="image", color=(0, 255, 0),
                     save_path=None):
    """
    Displays the image with the current_vision list of contours

    This can be used to display detected contours (object outlines) on an image (numpy array)

    .. code-block:: python

        import ovl
        import cv2

        image = cv2.imread("path/to/image.png")

        (define a vision pipeline to detect contours)

        contours, _ = vision.detect(image)

        ovl.display_contours(image, contours)


    :param image: image from which the contours were taken from, numpy array or image path
    :param contours: the list of contours to display_image
    :param color: the color of the of the contours outline
    :param display_loop: if display is used in a loop should be true, else false
    :param window_name: the window name, this is useful when display multiple different images at the same time
    :param amount: amount of contours to display
    :param save_path: if the image should be saved, pass the wanted result path
    :return: the image with the drawn contours.
    """
    if type(image) is str:
        image = open_image(image)
    image_for_display = copy.copy(image)
    if amount == math.inf:
        amount = 0
    contours = contours[0:amount] if amount > 0 else contours
    cv2.drawContours(image_for_display, contours, -1, color, 2)
    if type(save_path) is str:
        cv2.imwrite(save_path, image_for_display)
    display_image(image_for_display, window_name=window_name, display_loop=display_loop)
    return image_for_display
